"""
IPv6 Interface Data Type Class
"""

# pylint: disable=too-few-public-methods

import ipaddress

import nadap.mixin.ip
import nadap.types.base

DOC_DT_NAME = "IPv6 Interface"
DOC_DT_DESCRIPTION = """
An **ip6_interface** data type tests data for being an instance of
python's built-in class `str` and the string represents an IPv6
interface.
"""
DOC_DT_FEATURES = """
- Validate IP specific attributes
- Validate against allowed and not allowed IPv6 interfaces
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: ip6_interface
description: "Example IPv6 interface definition"

multicast: False
private: True

not_allowed_values:
  - 2001:cafe::1/64
  - 2001:abba::1/64

reference:
  key: ref_key
"""


class Ip6Interface(nadap.mixin.ip.IpMixin, nadap.types.base.BaseType):
    """
    Add handling of strings representing IPv6 interfaces
    """

    _ip_data_type_name = "IPv6 interface"

    def _get_ip_obj(self, data):
        try:
            ipaddress.IPv6Address(data)
        except ValueError:
            try:
                ipaddress.IPv6Network(data)
            except ValueError:
                return ipaddress.IPv6Interface(data)
        raise ValueError()


DOC_DT_CLASS = Ip6Interface  # pylint: disable=invalid-name



"""
MAC address data type class
"""

# pylint: disable=too-few-public-methods

import re

import nadap.results
import nadap.types.base
import nadap.mixin.ranges
import nadap.mixin.allowed_value
import nadap.mixin.not_allowed_value
import nadap.references
import nadap.schema
from nadap.base import ValEnv, CONVERT_DATA, str_list_out
from nadap.errors import SchemaDefinitionError, DataValidationError


DOC_DT_NAME = "MAC Adress"
DOC_DT_DESCRIPTION = """
A **mac_address** data type tests data for being an instance of
python's built-in class `str` and if it represents a MAC address (EUI-48).
"""
DOC_DT_FEATURES = """
- Validate MAC notation
- Validate MAC format
- Supports MAC format conversion
- Validate MAC specific attributes
- Validate against allowed and not allowed MAC addresses
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = r"""
type: mac_address
description: "Example mac_address definition"
default_value: "00:00:02:aa:bb:cc"

notation: canonical
formats:
  - EUI_48
  - dot-separated
uaa: true
group: false
convert_notation: bit-reversed
convert_to_format: colon-separated

not_allowed_values:
  - 04-01-01-ff-ff-ff

reference: ref_key
"""

CANONICAL = "canonical"
BIT_REVERSED = "bit-reversed"

EUI_48_FORMAT = "EUI-48"  # aa-bb-cc-dd-ee-ff
COLON_FORMAT = "colon-separated"  # aa:bb:cc:dd:ee:ff
DOT_FORMAT = "dot-separated"  # aabb.ccdd.eeff
HEX12_FORMAT = "HEX-12"  # aabbccddeeff


class MacAddress(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.types.base.BaseType,
):
    """
    MAC address datatype class
    """

    data_type_name = "mac_address"
    _cls_python_classes = [str]
    _convert_to_classes = {}
    _doc_data_type = "MAC address"

    def __init__(self, **kwargs):
        self._notation = CANONICAL
        self._formats = []
        self._detected_format = None
        self._uaa = None
        self._group = None
        self._convert_notation = None
        self._convert_to_format = None
        super().__init__(**kwargs)

    @staticmethod
    def _check_format_str(format_, path):
        if format_ not in [
            EUI_48_FORMAT,
            COLON_FORMAT,
            DOT_FORMAT,
            HEX12_FORMAT,
        ]:
            raise SchemaDefinitionError(  # pylint: disable=raise-missing-from
                msg="Format unknown",
                path=path,
            )

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)

        if self._notation not in [CANONICAL, BIT_REVERSED]:
            raise SchemaDefinitionError(
                msg="Notation unknown",
                path=f"{schema_path}.notation",
            )
        nadap.schema.is_list(self._formats, f"{schema_path}.formats")
        for index, format_ in enumerate(self._formats):
            self._check_format_str(format_, f"{schema_path}.formats[{index}]")
        self._formats = sorted(self._formats)

        if self._convert_to_format is not None:
            self._check_format_str(
                self._convert_to_format, f"{schema_path}.convert_to_format"
            )

    def _pop_options(self, definition: dict, schema_path: str):
        self._notation = definition.pop("notation", self._notation)
        if "format" in definition:
            if "formats" in definition:
                raise SchemaDefinitionError(
                    msg="excludes key 'formats'",
                    path=f"{schema_path}.format",
                )
            _format = definition.pop("format")
            nadap.schema.is_str(_format, f"{schema_path}.format")
            self._check_format_str(_format, f"{schema_path}.format")
            self._formats = [_format]
        else:
            self._formats = definition.pop("formats", self._formats)
        self._uaa = definition.pop("uaa", self._uaa)
        self._group = definition.pop("group", self._group)
        self._convert_notation = definition.pop(
            "convert_notation", self._convert_notation
        )
        self._convert_to_format = definition.pop(
            "convert_to_format", self._convert_to_format
        )
        super()._pop_options(definition, schema_path)

    def _validate_scope(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        # pylint: disable=too-many-branches
        data = self._int2hex8(data)
        msg = ""
        messages = []
        if self._uaa is not None:
            if self._notation == CANONICAL:
                if self._uaa and data[1] in "2367abef":
                    msg = "MAC address is locally administered"
                    messages.append(msg)
                if not self._uaa and data[1] in "014589cd":
                    msg = "MAC address is universally administered"
                    messages.append(msg)
            else:  # bit-reversed (i.e. Token Ring)
                if self._uaa and data[0] in "4567cdef":
                    msg = "MAC address is locally administered"
                    messages.append(msg)
                if not self._uaa and data[0] in "012389ab":
                    msg = "MAC address is universally administered"
                    messages.append(msg)
        if self._group is not None:
            if self._notation == CANONICAL:
                if self._group and data[1] in "02468ace":
                    msg = "MAC address is a unicast address"
                    messages.append(msg)
                if not self._group and data[1] in "13579bdf":
                    msg = "MAC address is a group address"
                    messages.append(msg)
            else:  # bit-reversed (i.e. Token Ring)
                if self._group and data[0] in "01234567":
                    msg = "MAC address is a unicast address"
                    messages.append(msg)
                if not self._group and data[0] in "89abcdef":
                    msg = "MAC address is a group address"
                    messages.append(msg)
        if msg:
            if env is None:
                raise SchemaDefinitionError(
                    msg,
                    path,
                )
            for m in messages:
                env.findings.append(
                    nadap.results.ValidationFinding(
                        message=m,
                        path=path,
                    )
                )
            raise DataValidationError()

    def _detect_data_format(self, data) -> str:
        if "-" in data:
            return EUI_48_FORMAT
        if ":" in data:
            return COLON_FORMAT
        if "." in data:
            return DOT_FORMAT
        return HEX12_FORMAT

    def _apply_data_format(self, data: "any", fmt: str = None) -> "any":
        data = self._int2hex8(data)
        if fmt == EUI_48_FORMAT:
            return f"{data[0:2]}-{data[2:4]}-{data[4:6]}-{data[6:8]}-{data[8:10]}-{data[10:12]}"
        if fmt == COLON_FORMAT:
            return f"{data[0:2]}:{data[2:4]}:{data[4:6]}:{data[6:8]}:{data[8:10]}:{data[10:12]}"
        if fmt == DOT_FORMAT:
            return f"{data[0:4]}.{data[4:8]}.{data[8:12]}"
        return data

    def _validate_format(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> str:
        # pylint: disable=too-many-branches
        if not self._formats:
            format_mismatch = "String does not represent a MAC address"
        else:
            format_mismatch = (
                "String does not represent a MAC address in format "
                + str_list_out(self._formats)
            )
        ret_value = None
        try:
            if "-" in data:
                # EUI-48 candidate
                if self._formats and EUI_48_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                if not re.fullmatch("([0-9a-f]{2}-){5}[0-9a-f]{2}", data):
                    raise ValueError(format_mismatch)
                self._detected_format = EUI_48_FORMAT
                ret_value = data.replace("-", "")
            elif ":" in data:
                # colon format candidate
                if self._formats and COLON_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                if not re.fullmatch(r"([0-9a-f]{2}\:){5}[0-9a-f]{2}", data):
                    raise ValueError(format_mismatch)
                self._detected_format = COLON_FORMAT
                ret_value = data.replace(":", "").lower()
            elif "." in data:
                # dotted format candidate
                if self._formats and DOT_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                if not re.fullmatch(r"([0-9a-f]{4}\.){2}[0-9a-f]{4}", data):
                    raise ValueError(format_mismatch)
                self._detected_format = DOT_FORMAT
                ret_value = data.replace(".", "").lower()
            elif not re.fullmatch("[0-9a-f]{12}", data):
                raise ValueError(format_mismatch)
            else:
                self._detected_format = HEX12_FORMAT
                ret_value = data
        except ValueError as e:
            if env is None:
                raise SchemaDefinitionError(  # pylint: disable=raise-missing-from
                    str(e),
                    path,
                )
            self._create_finding_with_error(str(e), path, env)
        return self._hex2int(ret_value)

    @staticmethod
    def _hex2int(h: str) -> int:
        return int(h, 16)

    @staticmethod
    def _int2hex8(i: int) -> str:
        h = hex(i)
        return (14 - len(h)) * "0" + h[2:]

    @classmethod
    def _apply_convert_notation(cls, data: str) -> str:
        def _translate_hex(h):
            bit_str = bin(int(h, 16))[2:]
            bit_str = (4 - len(bit_str)) * "0" + bit_str
            reversed_bits = bit_str[::-1]
            hex_ = hex(int(reversed_bits, 2))
            return hex_[-1]

        ret_mac = ""
        for byte in [
            data[0:2],
            data[2:4],
            data[4:6],
            data[6:8],
            data[8:10],
            data[10:12],
        ]:
            ret_mac += _translate_hex(byte[1]) + _translate_hex(byte[0])
        return ret_mac

    def _preprocess_data(self, data: any, env: "ValEnv"):
        if CONVERT_DATA in env.flags:
            if self._convert_notation is not None:
                data = self._int2hex8(data)
                data = self._apply_convert_notation(data)
                data = self._hex2int(data)
            if self._convert_to_format is not None:
                data = self._apply_data_format(data, self._convert_to_format)
            else:
                data = self._apply_data_format(data, self._detected_format)
        else:
            data = self._apply_data_format(data, self._detected_format)
        return data

    def _test_data_type(self, data: str, path: str, env: "ValEnv" = None) -> "str":
        data = super()._test_data_type(data=data, path=path, env=env).lower()
        data = self._validate_format(data=data, path=path, env=env)
        self._validate_scope(data=data, path=path, env=env)
        return data

    @property
    def restrictions(self) -> "list[str]":
        """
        Get all restrictions for valid data
        """
        ret_list = super().restrictions
        mac_restrictions = []
        if self._formats:
            mac_restrictions.append(
                ("must be in format " + str_list_out(self._formats))
            )
        if self._uaa is not None or self._group is not None:
            if self._uaa is not None:
                mac_restrictions.append(
                    f"must be {'universally' if self._uaa else 'locally'} administered"
                )
            if self._group is not None:
                mac_restrictions.append(
                    f"must be a {'group' if self._group else 'unicast'} address"
                )
            mac_restrictions.append(f"treated as in {self._notation} notation")

        if mac_restrictions:
            ret_list.extend(
                ["MAC address:"]
                + [f"{self._markdown_indent}- {r}" for r in mac_restrictions]
            )
        return ret_list

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        return self._doc_data_type

    @property
    def markdown_data_type(self) -> str:
        """Get markdown data type string"""
        return f"<code>{self._doc_data_type}</code>"

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        allowed_notations = (
            f"allowed_values:<br>{cls._markdown_indent}- "
            + f"{CANONICAL}<br>{cls._markdown_indent}- {BIT_REVERSED}"
        )
        allowed_formats = (
            f"allowed_values:<br>{cls._markdown_indent}- "
            + f"{EUI_48_FORMAT}<br>{cls._markdown_indent}- {COLON_FORMAT}"
            + f"<br>{cls._markdown_indent}- {DOT_FORMAT}<br>{cls._markdown_indent}- "
            + f"{HEX12_FORMAT}"
        )
        f_desc = (
            f"- {EUI_48_FORMAT} example: 01-01-01-aa-bb-cc<br>"
            + f"- {COLON_FORMAT} example: 01:01:01:aa:bb:cc<br>"
            + f"- {DOT_FORMAT} example: 0101.01aa.bbcc<br>"
            + f"- {HEX12_FORMAT} example: 010101aabbcc"
        )
        return super()._doc_options_md_upper_part() + [
            f"| **notation** | <code>enum</code> | {CANONICAL} | {allowed_notations} | "
            + "Canonical is used in i.e. Ethernet<br>Bit-reversed is used i.e. in Token Ring|",
            f"| **format** | <code>enum</code> | | {allowed_formats} | {f_desc} |",
            "| **formats** | <code>list[enum]</code> | | excludes: format | |",
            f"| &nbsp;&nbsp;- < format > | <code>enum</code> | | {allowed_formats} | {f_desc} |",
            "| **uaa** | <code>bool</code> | | | "
            + "MAC address must be universally (true) or locally (false) administered |",
            "| **group** | <code>bool</code> | | | "
            + "MAC address must be a group (true) or a unicast (false) address |",
            "| **convert_notation** | <code>bool</code> | <code>False</code> | | |",
            f"| **convert_to_format** | <code>enum</code> | | {allowed_formats} | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            f"notation: <{CANONICAL}|{BIT_REVERSED}>",
            f"format: <{EUI_48_FORMAT}|{COLON_FORMAT}|{DOT_FORMAT}|{HEX12_FORMAT}>",
            "formats:",
            f"  - <{EUI_48_FORMAT}|{COLON_FORMAT}|{DOT_FORMAT}|{HEX12_FORMAT}>",
            "uaa: <true|false>",
            "group: <true|false>",
            "convert_notation: <true|false>",
            f"convert_to_format: <{EUI_48_FORMAT}|{COLON_FORMAT}|{DOT_FORMAT}|{HEX12_FORMAT}>",
        ]


DOC_DT_CLASS = MacAddress  # pylint: disable=invalid-name



"""
Any type classes
"""

import nadap.types.base
import nadap.schema

DOC_DT_NAME = "Any"
DOC_DT_DESCRIPTION = """
An **any** data type represents any data.
"""
DOC_DT_FEATURES = """
- Supports **Referencing Feature**
- Data type conversion is not supported
"""
DOC_DT_YAML_EXAMPLE = """
type: any
description: "Example for an any definition"
default_value: ["nadap", "rulez"]

reference:
  key: ref_key
  mode: producer
  producer_scope: global
"""


class Any(nadap.types.base.BaseType):
    """
    Any datatype class
    """

    # pylint: disable=too-few-public-methods

    data_type_name = "any"
    _convert_to_classes = {}

    def _test_data_type(self, data: any, path: str, env=None):
        # pylint: disable=unused-argument
        return data


DOC_DT_CLASS = Any  # pylint: disable=invalid-name



"""
Execption classes
"""

# pylint: disable=too-few-public-methods


class SchemaDefinitionError(Exception):
    """
    Error while parsing schema definition.
    """

    def __init__(self, msg: str, path: str = None):
        if path is not None:
            msg = f"{path}: {msg}"
        super().__init__(msg)


class DataValidationError(Exception):
    """
    Exception indicating that data validation failed.
    """


class NadapReferenceError(Exception):
    """
    Class to identify an error during reference processing
    """



"""
Reference classes and errors
"""

# pylint: disable=too-few-public-methods

import nadap.results
from nadap.base import OPT
from nadap.errors import NadapReferenceError


class RefDef:
    """
    Represent all information from reference definition for a data type.
    """

    def __init__(
        self,
        ref_key: str,
        ref_options: OPT,
        ref_credits: int = None,
        ns_separator: str = None,
    ):
        self.key = ref_key
        self.options = ref_options
        if ref_credits is None and OPT.CONSUMER in ref_options:
            self.credits = 1
        else:
            self.credits = ref_credits
        self.ns_separator = ns_separator

    def __eq__(self, v: "RefDef"):
        return (
            self.key == v.key
            and self.options == v.options
            and self.credits == v.credits
            and self.ns_separator == v.ns_separator
        )


class ReferenceElement:
    """ReferenceElement stores all information regarding a reference value analyzed in data"""

    def __init__(
        self,
        ref_def: RefDef,
        path: str = "",
        value: any = None,
        namespace: str = "",
    ) -> None:
        """
        Args:
            ref_def: RefDef object - Reference definition from schema

            path: string - Value's path in data structure

            value: any - Value data

            namespace: string - active namepsace during reference creation

        Raises:
            ValueError: If arg's data type is wrong.
        """
        self.ref_def = ref_def
        self.namespace = namespace
        self.path = path
        self.value = value
        self.consumes_from = []
        self.provides_to = []

    @property
    def key(self) -> str:
        """
        Get reference key from reference definition
        """
        return self.ref_def.key

    @property
    def options(self) -> str:
        """
        Get reference options from reference definition
        """
        return self.ref_def.options

    @property
    def credits(self) -> str:
        """
        Get reference credits from reference definition
        """
        return self.ref_def.credits


class ConsumerElement(ReferenceElement):
    """
    Sub-class to ReferenceElement
    Specialized for Consumer References
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments

    def __init__(
        self,
        ref_def: RefDef,
        path: str = "",
        value: any = None,
        namespace: str = "",
        provider_namespace: str = "",
    ) -> None:
        """
        Args:
            < ReferenceElement's arguements >

            provider_namespace: string - namespace where to search
                                for providers (namespace lookup feature)

        Raises:
            ValueError: If arg's data type is wrong.
        """
        super().__init__(
            ref_def=ref_def,
            path=path,
            value=value,
            namespace=namespace,
        )
        self.provider_namespace = provider_namespace


class ReferenceFinding(nadap.results.Finding):
    """
    Sub-class to Finding
    ReferenceFinding stores all information to a reference issue and
    provides a method to present all information as a string (makes it easy for user interfaces)
    """

    def __init__(
        self, message: str, path: str = None, reference: ReferenceElement = None
    ) -> None:
        super().__init__(message, path)
        self.reference = reference

    def __str__(self):
        ns_lookup = False
        namespace = ""
        if self.reference:
            if (
                isinstance(self.reference, ConsumerElement)
                and self.reference.provider_namespace
            ):
                ns_lookup = True
                namespace = self.reference.namespace + " > "
            elif self.reference.namespace:
                namespace = self.reference.namespace + " > "
        path = self.path + ": " if self.path else ""
        message = (
            f"{self.message} in '{self.reference.provider_namespace}'"
            if ns_lookup
            else self.message
        )
        return namespace + path + message


class GlobalSpace:
    """
    GlobalSpace is a space where to register references in global scope.
    Adding References will stored according to the mode and
    links (consumer<>producer) to other refererences will be created.
    """

    uOPT = OPT.UNIQUE_GLOBAL
    pOPT = OPT.PRODUCER_GLOBAL
    cOPT = OPT.CONSUMER_GLOBAL

    def __init__(self) -> None:
        self.references = {}

    def get_uniques_by_value(self, value: any, ref_key: str) -> list[ReferenceElement]:
        """
        Get all unique elements to a reference key with given value
        """
        return self._get_elements_by_value(value, ref_key, "uniques")

    def get_producers_by_value(
        self, value: any, ref_key: str
    ) -> list[ReferenceElement]:
        """
        Get all producer elements to a reference key with given value
        """
        return self._get_elements_by_value(value, ref_key, "producers")

    def get_consumers_by_value(
        self, value: any, ref_key: str
    ) -> list[ReferenceElement]:
        """
        Get all consumer elements to a reference key with given value
        """
        return self._get_elements_by_value(value, ref_key, "consumers")

    def _get_elements_by_value(
        self, value: any, ref_key: str, _list: str
    ) -> list[ReferenceElement]:
        if ref_key not in self.references:
            return []
        return [e for e in self.references[ref_key][_list] if e.value == value]

    def get_uniques_by_ref_key(self, ref_key: str) -> list[ReferenceElement]:
        """
        Get all unique elements to a reference key
        """
        return self.references.get(ref_key, {}).get("uniques", [])

    def get_producers_by_ref_key(self, ref_key: str) -> list[ReferenceElement]:
        """
        Get all producer elements to a reference key
        """
        return self.references.get(ref_key, {}).get("producers", [])

    def get_consumers_by_ref_key(self, ref_key: str) -> list[ReferenceElement]:
        """
        Get all consumer elements to a reference key
        """
        return self.references.get(ref_key, {}).get("consumers", [])

    def get_unique_values_by_ref_key(self, reference_key: str) -> set:
        """
        Returns:
            list - with all 'unique' values with the same reference key
        """
        unique_values = set()
        for i in self.get_uniques_by_ref_key(reference_key):
            unique_values.add(i.value)
        return unique_values

    def _add_ref_key(self, reference_key: str):
        """
        If reference key is new to this space,
        creates all required lists (unique,producers and consumers) for key 'ref_key' in references.

        Args:
            ref_key: string - reference key from schema definition
        """
        if reference_key in self.references:
            return
        self.references[reference_key] = {
            "uniques": [],
            "producers": [],
            "consumers": [],
            "producer_credits": {},
            "consumer_credits": {},
        }

    def _add_unique_element(self, e: ReferenceElement):
        if match := self.get_uniques_by_value(e.value, e.key):
            raise NadapReferenceError(f"Reference already defined at {match[0].path}")
        self.references[e.key]["uniques"].append(e)

    def _add_producing_element(self, e: ReferenceElement) -> list[ReferenceElement]:
        self.references[e.key]["producers"].append(e)
        # Check if value can be used as a key in a dictionary.
        # If not convert to string.
        try:
            hash(e.value)
            value = e.value
        except TypeError:
            value = str(e.value)
        if value in self.references[e.key]["producer_credits"]:
            if e.credits is None:
                # Producer sets credits for this value to infinite (None)
                self.references[e.key]["producer_credits"][value] = None
            elif self.references[e.key]["producer_credits"][value] is not None:
                # Only add credits if not a producer created
                # this value with infinite credits (None)
                self.references[e.key]["producer_credits"][value] += e.credits
        else:
            # New produced value. Just set credits.
            self.references[e.key]["producer_credits"][value] = e.credits
        return self._link_to_consumers(e)

    def _add_consuming_element(self, e: ReferenceElement) -> list[ReferenceElement]:
        self.references[e.key]["consumers"].append(e)
        # Check if value can be used as a key in a dictionary.
        # If not convert to string.
        try:
            hash(e.value)
            value = e.value
        except TypeError:
            value = str(e.value)
        # Defaulting credits to 1.
        new_credits = 1 if e.credits is None else e.credits
        if value in self.references[e.key]["consumer_credits"]:
            # Add credits to existing credits for this value
            self.references[e.key]["consumer_credits"][value] += new_credits
        else:
            # Just set credits for this value
            self.references[e.key]["consumer_credits"][value] = new_credits
        return self._link_to_producers(e)

    def add_element(self, e: ReferenceElement) -> list[ReferenceElement]:
        """
        Add a new reference to the space, if reference options matches the class scope:
        - stores reference according to mode in corresponding lists under the reference key
        - create all links (consumer<>producer) to other references in this space

        Args:
            e: ReferenceElement object - New reference for this space

        Returns:
            list of ReferenceElements - which were linked to the new reference

        Raises:
            NadapReferenceError: If reference already in this space
        """
        self._add_ref_key(e.key)
        if self.uOPT in e.options:
            self._add_unique_element(e)
        if self.pOPT in e.options:
            return self._add_producing_element(e)
        if self.cOPT in e.options:
            return self._add_consuming_element(e)
        return []

    def _link_to_producers(self, consumer: ReferenceElement) -> list[ReferenceElement]:
        producer_list = []
        for producer in self.references[consumer.key]["producers"]:
            if (
                consumer.value == producer.value
                and producer not in consumer.consumes_from
            ):
                # Maybe linking already from another NS or in Global
                consumer.consumes_from.append(producer)
                producer.provides_to.append(consumer)
                producer_list.append(producer)
        return producer_list

    def _link_to_consumers(self, producer: ReferenceElement) -> list[ReferenceElement]:
        consumer_list = []
        for consumer in self.references[producer.key]["consumers"]:
            if (
                producer.value == consumer.value
                and consumer not in producer.provides_to
            ):
                # Maybe linking already from another NS or in Global
                producer.provides_to.append(consumer)
                consumer.consumes_from.append(producer)
                consumer_list.append(consumer)
        return consumer_list

    def get_producer_consumer_issues(self) -> list[ReferenceFinding]:
        """
        Collects and returns
        - all consumer references without a linked provider and
        - all provider references without a linked consumer, if orphan procuders are not allowed,
        as a list of ReferenceFindings

        Returns:
            list of ReferenceFindings
        """
        results = []
        for ref_lists in self.references.values():
            for consumer in ref_lists["consumers"]:
                if len(consumer.consumes_from) == 0:
                    # No producer found during analysis!
                    results.append(
                        ReferenceFinding(
                            path=consumer.path,
                            message="No producer found",
                            reference=consumer,
                        )
                    )
            for producer in ref_lists["producers"]:
                if (
                    OPT.ALLOW_ORPHAN_PRODUCER not in producer.options
                    and len(producer.provides_to) == 0
                ):
                    # Producer has no consumer!
                    results.append(
                        ReferenceFinding(
                            path=producer.path,
                            message="Producer has no consumer",
                            reference=producer,
                        )
                    )
        return results

    def get_credit_issues(self) -> list[ReferenceFinding]:
        """
        Checks if consumer credits exceeds provider credits.

        Returns:
            list of ReferenceFindings
        """
        results = []
        for ref_key, ref_lists in self.references.items():
            for producer_value, producer_credits in ref_lists[
                "producer_credits"
            ].items():
                if producer_credits is not None:
                    consumer_credits = ref_lists["consumer_credits"].get(
                        producer_value, None
                    )
                    if (
                        consumer_credits is not None
                        and consumer_credits > producer_credits
                    ):
                        results.append(
                            ReferenceFinding(
                                path="n.a.",
                                message=f"reference key '{ref_key}': Consumer credits exceeds "
                                + f"producer credits for value '{producer_value}'",
                            )
                        )
        return results


class NameSpace(GlobalSpace):
    """
    Sub-class to GlobalSpace
    Namespace is a space in namespace scope
    """

    uOPT = OPT.UNIQUE
    pOPT = OPT.PRODUCER
    cOPT = OPT.CONSUMER


class References:
    """
    References stores all references according to the current active namespace and
    according to the reference options (mode, scope, ...)
    """

    def __init__(self, namespace: str = "") -> None:
        self.namespace = namespace
        self.globalspace_obj = GlobalSpace()
        self.ns_obj_store = {}
        self._create_namespace(namespace)
        self.namespace_obj = self.ns_obj_store[namespace]

    def change_namespace(self, namespace: str) -> None:
        """
        Change namespace
        """
        self._create_namespace(namespace)
        self.namespace_obj = self.ns_obj_store[namespace]
        self.namespace = namespace

    def _create_namespace(self, namespace: str) -> None:
        if namespace not in self.ns_obj_store:
            self.ns_obj_store[namespace] = NameSpace()

    def add_element(self, reference: ReferenceElement) -> list[ReferenceElement]:
        """
        Add a new reference to the reference store, if reference options matches the class scope:
        - add the reference to global space and
        - if mode is consumer and provider namespace is given (namespace lookup),
          add the reference to provider namespace,
          else add the reference to the current active namespace

        Args:
            reference: ReferenceElement object - New reference for this store

        Returns:
            list of ReferenceElements - which were linked to the new reference

        Raises:
            Cd2tReferenceError: If reference already in this store
        """
        reference.namespace = self.namespace
        linked_elements = self.globalspace_obj.add_element(reference)

        if isinstance(reference, ConsumerElement) and reference.provider_namespace:
            self._create_namespace(reference.provider_namespace)
            linked_elements += self.ns_obj_store[
                reference.provider_namespace
            ].add_element(reference)
        else:
            linked_elements += self.namespace_obj.add_element(reference)
        return linked_elements

    def get_producer_consumer_issues(self):
        """
        Collects and returns
        - all consumer references without a linked provider and
        - all provider references without a linked consumer, if orphan procuders are not allowed,
        from all namespaces as a list of ReferenceFindings.
        Note: As each reference in global space is also in a namespace, globalspace is skipped.

        Returns:
            list of ReferenceFindings
        """
        credit_results = self.globalspace_obj.get_credit_issues()
        for finding in credit_results:
            finding.message = "Global " + finding.message
        results = credit_results

        for namespace, ns_obj in self.ns_obj_store.items():
            results += ns_obj.get_producer_consumer_issues()
            credit_results = ns_obj.get_credit_issues()
            for finding in credit_results:
                finding.message = f"Namespace '{namespace}' {finding.message}"
            results += credit_results

        return results

    def same_unique(self, reference: ReferenceElement) -> list[ReferenceElement]:
        """
        Returns first found 'unique' references with
        the same reference key and value as given reference.

        Args:
            reference: ReferenceElement

        Returns:
            list of ReferenceElements
        """
        others = self.namespace_obj.get_uniques_by_value(reference.value, reference.key)
        if OPT.UNIQUE_GLOBAL in reference.options:
            others += self.globalspace_obj.get_uniques_by_value(
                reference.value, reference.key
            )
        if others:
            return others[0]
        return None

    def get_unique_values_by_ref_key(self, reference_key: str) -> set:
        """
        Returns all found 'unique' values with the same reference key.

        Args:
            reference_key: string

        Returns:
            list of values
        """
        ns_uniques = set(self.namespace_obj.get_unique_values_by_ref_key(reference_key))
        global_unique = set(
            self.globalspace_obj.get_unique_values_by_ref_key(reference_key)
        )
        return ns_uniques.union(global_unique)



"""
Bool type classes
"""

# pylint: disable=too-few-public-methods

from nadap.base import UNDEFINED
import nadap.types.bool

DOC_DT_NAME = "Bool-True"
DOC_DT_DESCRIPTION = """
A **bool_true** data type tests data for being an instance of
python's built-in class `bool`.
It has a pre-defined default value set to `True`.
"""
DOC_DT_FEATURES = """
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: bool_true
description: "Example for a bool_true definition"

reference:
  key: ref_key
  mode: producer
  producer_scope: global
"""


class BoolTrue(nadap.types.bool.Bool):
    """
    Bool datatype class with true as default value
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.default_value is UNDEFINED:
            self.default_value = True


DOC_DT_CLASS = BoolTrue  # pylint: disable=invalid-name



"""
Hostname or IP address data type class
"""

# pylint: disable=too-few-public-methods

import nadap.types.base
import nadap.types.hostname
import nadap.types.ip_address
import nadap.types.ip4_address
import nadap.types.ip6_address
import nadap.references
import nadap.schema
from nadap.base import ValEnv
from nadap.errors import DataValidationError, SchemaDefinitionError

DOC_DT_NAME = "Hostname or IP Adress "
DOC_DT_DESCRIPTION = """
A **hostname_or_ip** data type tests data for being an instance of
python's built-in class `str` and if it represents a hostname/FQDN
or an IPv4/IPv6 address.
"""
DOC_DT_FEATURES = (
    """
Hostname Features:
"""
    + nadap.types.hostname.DOC_DT_FEATURES
    + """
IP Address Features:
"""
    + nadap.types.ip_address.DOC_DT_FEATURES
)
DOC_DT_YAML_EXAMPLE = r"""
type: hostname_or_ip
description: "Example hostname_or_ip definition"
default_value: "my-host"

hostname:
    minimum: 3
    maximum: 16
    minimum_labels: 3
    maximum_labels: 7

    not_allowed_values:
    - nadap-.*
    relax_length: false

ip:
    version: 4
    multicast: False
    private: True

    not_allowed_values:
    - 10.10.99.1
    - 192.168.0.1

reference: ref_key
"""


class HostnameOrIP(nadap.types.base.BaseType):
    """
    Hostname or IP Address datatype class
    """

    data_type_name = "hostname_or_ip"
    _cls_python_classes = [str]
    _convert_to_classes = {}
    _support_replace_empty_to = True
    _doc_data_type = "str"

    def __init__(self, **kwargs):
        self._ip_definition = {}
        self._ip_version = None
        self._hostname_definition = {}
        self._ip_dt_obj = None
        self._hostname_dt_obj = None
        super().__init__(**kwargs)

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        ip_path = f"{schema_path}.ip"
        nadap.schema.is_dict(self._ip_definition, ip_path)
        self._ip_version = self._ip_definition.pop("version", None)
        if self._ip_version:
            nadap.schema.is_int(self._ip_version, f"{ip_path}.version")
            if self._ip_version not in [4, 6]:
                raise SchemaDefinitionError(
                    msg="Must be either 4 or 6", path=f"{ip_path}.version"
                )
            if self._ip_version == 4:
                ip_dt_class = nadap.types.ip4_address.Ip4Address
            else:
                ip_dt_class = nadap.types.ip6_address.Ip6Address
        else:
            ip_dt_class = nadap.types.ip_address.IpAddress
        self._ip_dt_obj = ip_dt_class(
            schema=self._schema,
            definition=self._ip_definition,
            schema_path=ip_path,
        )
        self._hostname_dt_obj = nadap.types.hostname.Hostname(
            schema=self._schema,
            definition=self._hostname_definition,
            schema_path=f"{schema_path}.hostname",
        )

    def _pop_options(self, definition: dict, schema_path: str):
        ref_definition = definition.get("reference", {})
        ref_definitions = definition.get("references", {})
        self._ip_definition = definition.pop("ip", self._ip_definition)
        self._hostname_definition = definition.pop(
            "hostname", self._hostname_definition
        )
        if ref_definition:
            self._ip_definition["reference"] = ref_definition
            self._hostname_definition["reference"] = ref_definition
        if ref_definitions:
            self._ip_definition["references"] = ref_definitions
            self._hostname_definition["references"] = ref_definitions
        super()._pop_options(definition, schema_path)

    def validate(
        self,
        data: any,
        path: str,
        env: "ValEnv",
    ) -> "any":
        """
        - Check if data matches IP Data type
        - If fails, check if data matches Hostname Data Type
        - apply preprocessing and perform referencing
        """
        self._test_data_type(data=data, path=path, env=env)
        len_findings = len(env.findings)
        try:
            data = self._ip_dt_obj.validate(data=data, path=path, env=env)
        except DataValidationError:
            try:
                data = self._hostname_dt_obj.validate(data=data, path=path, env=env)
                if len_findings < len(env.findings):
                    del env.findings[len_findings - len(env.findings) :]
            except DataValidationError:
                env.findings.insert(
                    len_findings,
                    nadap.results.ValidationFinding(
                        message="Data matches wether defined hostname nor defined IP address",
                        path=path,
                    ),
                )
                raise DataValidationError()  # pylint: disable=raise-missing-from
        return data

    @property
    def restrictions(self) -> "list[str]":
        """
        Get all restrictions for valid data
        """
        ip_restrictions = self._ip_dt_obj.restrictions
        hostname_restrictions = self._hostname_dt_obj.restrictions
        ret_list = []
        if self._ip_version or ip_restrictions:
            ret_list.append("IP Address Restrictions:")
            if self._ip_version:
                ret_list.append(
                    f"{self._markdown_indent}IP Version: {self._ip_version}"
                )
            ret_list.extend([(self._markdown_indent + x) for x in ip_restrictions])
        if hostname_restrictions:
            ret_list.append("Hostname Restrictions:")
            ret_list.extend(
                [(self._markdown_indent + x) for x in hostname_restrictions]
            )
        return ret_list

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **hostname** | <code>dict</code> | | | Hostname data type definition part |",
            "| &nbsp;&nbsp;**minimum** | <code>int</code> | | min: 1 | Data's length must be "
            + "greater or equal |",
            "| &nbsp;&nbsp;**maximum** | <code>int</code> | | min: 1<br>>= 'minimum' | Data's "
            + "length must be lower or equal |",
            "| &nbsp;&nbsp;**replace_empty_to** | <code>any</code> | | |  An empty input value "
            + "will be replaced by this data |",
            "| &nbsp;&nbsp;**relax_length** | <code>bool</code> | <code>True</code> | "
            + "| Relax hostname length "
            + "from 63 to 255 characters |",
            "| &nbsp;&nbsp;**minimum_labels** | <code>int</code> | | min: 1 | Minimum amount of "
            + "labels within in hostname |",
            "| &nbsp;&nbsp;**maximum_labels** | <code>int</code> | | min: 1<br>>= 'minimum_labels' "
            + "| Maximum amount of labels within in hostname |",
            "| &nbsp;&nbsp;**allowed_values** | <code>list[str]</code> | | |  Data must match one "
            + "of these regex patterns |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;- < str > | <code>str</code> | | | |",
            "| &nbsp;&nbsp;**not_allowed_values** | <code>list[str]</code> | | |  Data mustn't "
            + "match any of these regex patterns |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;- < str > | <code>str</code> | | | |",
            "| **ip** | <code>dict</code> | | | IP address data type definition part |",
            "| **allowed_ranges** | <code>list[dict]</code> | | min length: 1 | "
            + "Value must be within defined ranges |",
            f"| {cls._markdown_indent}- **start** | <code>str</code> | | required |",
            f"| {cls._markdown_indent * 2}**end** | <code>str</code> | | required |",
            "| **not_allowed_ranges** | <code>list[dict]</code> | | "
            + "| Value mustn't be within defined ranges |",
            f"| {cls._markdown_indent}- **start** | <code>str</code> | | required |",
            f"| {cls._markdown_indent * 2}**end** | <code>str</code> | | required |",
            "| &nbsp;&nbsp;**allowed_values** | <code>list</code> | | |  Data must match one of "
            + "these values |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;- < value > | <code>any</code> | "
            + "| Must match data type's type(s) | |",
            "| &nbsp;&nbsp;**not_allowed_values** | <code>list</code> | "
            + "| |  Data mustn't match all of these "
            + "values |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;- < value > | <code>any</code> | "
            + "| Must match data type's type(s) | |",
            "| &nbsp;&nbsp;**link_local** | <code>bool</code> | | | IP must be link-local |",
            "| &nbsp;&nbsp;**loopback** | <code>bool</code> | | | IP must be loopback |",
            "| &nbsp;&nbsp;**multicast** | <code>bool</code> | | | IP must be multicast |",
            "| &nbsp;&nbsp;**private** | <code>bool</code> | | | IP must be private |",
            "| &nbsp;&nbsp;**public** | <code>bool</code> | | | IP must be public/global |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "ip:",
            "  convert_to: <str>",
            "  allowed_ranges:",
            "    - start: <str>",
            "      end: <str>",
            "  not_allowed_ranges:",
            "   - start: <str>",
            "     end: <str>",
            "  allowed_values:",
            "  - <any>",
            "  not_allowed_values:",
            "  - <any>",
            "  link_local: <true|false",
            "  loopback: <true|false",
            "  multicast: <true|false",
            "  private: <true|false",
            "  public: <true|false",
            "hostname:",
            "  version: <int>",
            "  maximum: <int>",
            "  minimum: <int>",
            "  replace_empty_to: <any>",
            "  relax_length: <true|false>",
            "  minimum_labels: <int>",
            "  maximum_labels: <int>",
            "  allowed_values:",
            "  - <str>",
            "  not_allowed_values:",
            "  - <str>",
        ]


DOC_DT_CLASS = HostnameOrIP  # pylint: disable=invalid-name



"""
Basic constants and functions
"""

# pylint: disable=too-few-public-methods

import enum
import copy
import re

if hasattr(re, "NOFLAG"):
    NOFLAG = re.NOFLAG
else:
    NOFLAG = 0


class OPT(enum.IntFlag):  # pylint: disable=no-member
    """
    Flags used to define a reference
    """

    NONE = 1
    UNIQUE = 2
    UNIQUE_GLOBAL = 4
    PRODUCER = 8
    PRODUCER_GLOBAL = 16
    CONSUMER = 32
    CONSUMER_GLOBAL = 64
    ALLOW_ORPHAN_PRODUCER = 128


INIT_OPTIONS = OPT.UNIQUE | OPT.UNIQUE_GLOBAL | OPT.PRODUCER | OPT.PRODUCER_GLOBAL


class RegexObject:
    """
    Represents a regex allowed or not allowed value
    """

    def __init__(self, pattern: str, multiline: bool, fullmatch: bool):
        self.multiline = multiline
        self.fullmatch = fullmatch
        self.pattern = re.compile(pattern, self.flags)

    @property
    def flags(self) -> list:
        """
        Create a list of re args to pass to match
        """
        if self.multiline:
            return re.MULTILINE
        return NOFLAG

    def match(self, data: any) -> re.Match:
        """
        Test data if it matches
        Returns:
            re.Match object or None
        """
        if self.fullmatch:
            return self.pattern.fullmatch(data)
        return self.pattern.search(data)

    def __hash__(self):
        return self.pattern.pattern.__hash__()


class PreProcessingFlag(enum.IntFlag):  # pylint: disable=no-member
    """
    Flags used as pre-processing options
    """

    NOFLAG = 0
    SET_DEFAULTS = 1
    CONVERT_DATA = 2


class ValEnv:
    """
    Contains information and references required during
    validation and pre-processing
    """

    def __init__(
        self,
        references,
        findings,
        flags: PreProcessingFlag = PreProcessingFlag.NOFLAG,
    ):
        self.references = references
        self.findings = findings
        self.flags = flags


NOFLAG = PreProcessingFlag.NOFLAG
SET_DEFAULTS = PreProcessingFlag.SET_DEFAULTS
CONVERT_DATA = PreProcessingFlag.CONVERT_DATA
UNDEFINED = object()


def str_list_out(l: list) -> str:
    """
    creates a str from a list, representing an proper output
    """
    if len(l) == 0:
        return ""
    if len(l) == 1:
        return f"'{str(l[0])}'"
    return "'" + "', '".join([str(x) for x in l[:-1]]) + f"' or '{str(l[-1])}'"


def number_to_str_number(x: any) -> any:
    """
    Test if x is a str and represents an integer.
    If yes return a string 'x', else return x
    """
    if isinstance(x, str):
        try:
            int(x)
            x = f"'{x}'"
        except ValueError:
            try:
                float(x)
                x = f"'{x}'"
            except ValueError:
                pass
    return x


def merge_lists(left: list, right: list, list_merge: str = "append_rp") -> list:
    """
    Merge elements of list 'right' into list 'left'
    """
    if not isinstance(left, list):
        raise ValueError("'left' is not a list")
    if not isinstance(right, list):
        raise ValueError("'right' is not a list")
    _left = copy.deepcopy(left)
    if list_merge == "replace":
        return copy.deepcopy(right)
    if list_merge == "append":
        return _left + copy.deepcopy(right)
    if list_merge == "prepend":
        return copy.deepcopy(right) + _left
    if list_merge in ["append_rp", "prepend_rp"]:
        append = list_merge == "append_rp"
        for element in right:
            if element not in _left:
                if append:
                    _left.append(copy.deepcopy(element))
                else:
                    _left.insert(0, copy.deepcopy(element))
        return _left
    raise ValueError(f"'list_merge' option {list_merge} is not known")


def merge_dictionaries(
    left: dict, right: dict, recursive: bool = True, list_merge: str = "append_rp"
) -> dict:
    """
    Merge keys (recursively) from dictionary 'right' into dictionary 'left'.
    """
    if not isinstance(left, dict):
        raise ValueError("'left' is not a dictionary")
    if not isinstance(right, dict):
        raise ValueError("'right' is not a dictionary")
    _left = copy.deepcopy(left)
    for key, r_value in right.items():
        if key not in _left or not isinstance(r_value, type(_left[key])):
            # right element not in left or mismatching types => Just copy right to left
            _left[key] = copy.deepcopy(r_value)
        else:
            # left value is same type as right value
            if isinstance(r_value, list):
                # Merge left list with right list
                _left[key] = merge_lists(
                    left=_left[key], right=right[key], list_merge=list_merge
                )
            elif isinstance(r_value, dict):
                if recursive:
                    # Merge left dict with right dict
                    _left[key] = merge_dictionaries(
                        left=_left[key],
                        right=right[key],
                        recursive=recursive,
                        list_merge=list_merge,
                    )
                else:
                    # Overwrite left with right value
                    _left[key] = copy.deepcopy(r_value)
            else:
                # Something like integer or string and just copy right to left
                _left[key] = copy.deepcopy(r_value)
    return _left


def merge_template_data(
    template_data: dict,
    additional_data: dict,
    recursive: bool = True,
    list_merge: str = "append_rp",
) -> dict:
    """
    Merge definition from 'additional' template into template.
    """
    template_data = template_data.copy()
    template_data_keys = set(list(template_data.keys()))
    additional_data_keys = set(list(additional_data.keys()))
    return_dict = template_data
    for key in template_data_keys.intersection(additional_data_keys):
        if isinstance(additional_data[key], list):
            if isinstance(template_data[key], list):
                return_dict[key] = merge_lists(
                    left=template_data[key],
                    right=additional_data[key],
                    list_merge=list_merge,
                )
            else:
                return_dict[key] = additional_data[key]
        elif not isinstance(additional_data[key], dict) or not isinstance(
            template_data[key], dict
        ):
            return_dict[key] = additional_data[key]
        else:
            return_dict[key] = merge_dictionaries(
                left=template_data[key],
                right=additional_data[key],
                recursive=recursive,
                list_merge=list_merge,
            )
    for key in additional_data_keys - template_data_keys:
        return_dict[key] = additional_data[key]
    return return_dict



"""
RegexModeMixin
"""

# pylint: disable=too-few-public-methods


class RegexModeMixin:
    """
    Add options for regex mode
    """

    def __init__(self, **kwargs):
        self.regex_mode = False
        self.regex_fullmatch = True
        self.regex_multiline = False
        super().__init__(**kwargs)

    def _pop_options(self, definition: dict, schema_path: str):
        self.regex_fullmatch = definition.pop("regex_fullmatch", True)
        self.regex_multiline = definition.pop("regex_multiline", False)
        self.regex_mode = definition.pop("regex_mode", False)
        super()._pop_options(definition=definition, schema_path=schema_path)

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **regex_mode** | <code>bool</code> | <code>False</code> | | "
            + " Enable testing strings against regex strings |",
            "| **regex_fullmatch** | <code>bool</code> | <code>True</code> | | "
            + " Full string must match |",
            "| **regex_multiline** | <code>bool</code> | <code>False</code> | | "
            + " Regex *^* and *$* matches to newline character |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "regex_mode: <true|false>",
            "regex_fullmatch: <true|false>",
            "regex_multiline: <true|false>",
        ]



"""
MAC address data type class
"""

# pylint: disable=too-few-public-methods

import nadap.results
import nadap.types.base
import nadap.mixin.ranges
import nadap.mixin.allowed_value
import nadap.mixin.not_allowed_value
import nadap.schema
from nadap.base import ValEnv, CONVERT_DATA
from nadap.errors import SchemaDefinitionError, DataValidationError


DOC_DT_NAME = "BGP AS Number"
DOC_DT_DESCRIPTION = """
A **bgp_as** data type tests data for being an instance of
python's built-in class `str` or `int` and if it represents a BGP AS number.
"""
DOC_DT_FEATURES = """
- Validate BGP AS format
- Supports BGP AS format conversion
- Validate BGP AS specific attributes
- Validate against allowed and not allowed BGP AS numbers
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = r"""
type: bgp_as
description: "Example bgp_as definition"
default_value: 100111

format: asplain
public: false
4_byte: true
convert_to_format: asdot

not_allowed_values:
  - 100222

not_allowed_ranges:
  - start: 123000
    end: 125000

reference: ref_key
"""


ASDOT_FORMAT = "asdot"  # "65000.100"
ASPLAIN_FORMAT = "asplain"  # 11223344
STRING_FORMAT = "string"  # "11223344"
ASDOT_UNCLEAR = "asdot_unclear"


class BgpAs(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.types.base.BaseType,
):
    """
    BGP AS datatype class
    """

    data_type_name = "bgp_as"
    _cls_python_classes = [int, str]
    _convert_to_classes = {}
    _doc_data_type = "BGP AS number"

    def __init__(self, **kwargs):
        self._format = None
        self._detected_format = None
        self._public = None
        self._convert_to_format = None
        super().__init__(**kwargs)

    @staticmethod
    def _check_format_str(format_, path):
        if format_ not in [ASDOT_FORMAT, ASPLAIN_FORMAT, STRING_FORMAT]:
            raise SchemaDefinitionError(  # pylint: disable=raise-missing-from
                msg="Format unknown",
                path=path,
            )

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)

        if self._format is not None and self._format not in [
            ASDOT_FORMAT,
            ASPLAIN_FORMAT,
            STRING_FORMAT,
        ]:
            raise SchemaDefinitionError(
                msg="Notation unknown",
                path=f"{schema_path}.notation",
            )

        if self._convert_to_format is not None:
            self._check_format_str(
                self._convert_to_format, f"{schema_path}.convert_to_format"
            )

    def _pop_options(self, definition: dict, schema_path: str):
        if "format" in definition:
            _format = definition.pop("format")
            nadap.schema.is_str(_format, f"{schema_path}.format")
            self._check_format_str(_format, f"{schema_path}.format")
            self._format = _format
        self._public = definition.pop("public", self._public)
        self._convert_to_format = definition.pop(
            "convert_to_format", self._convert_to_format
        )
        super()._pop_options(definition, schema_path)

    @staticmethod
    def _raise_exceptions(messages: "list[str]", path: str, env: "ValEnv|None"):
        if messages:
            if isinstance(messages, str):
                messages = [messages]
            if env is None:
                raise SchemaDefinitionError(
                    messages[0],
                    path,
                )
            for m in messages:
                env.findings.append(
                    nadap.results.ValidationFinding(
                        message=m,
                        path=path,
                    )
                )
            raise DataValidationError()

    def _validate_scope(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        msg = None
        if self._public is not None:
            if self._public:
                if 64512 <= data <= 65534 or 4200000000 <= data <= 4294967294:
                    msg = "BGP AS number is for private use"
            elif data < 64512 or (
                data > 65534 and (data < 4200000000 or data > 4294967294)
            ):
                msg = "BGP AS number is for public use"
        self._raise_exceptions(messages=msg, path=path, env=env)

    def _detect_data_format(self, data) -> str:
        if isinstance(data, int):
            return ASPLAIN_FORMAT
        if isinstance(data, str):
            if "." in data:
                return ASDOT_FORMAT
            try:
                if int(data) <= 65535:
                    return ASDOT_UNCLEAR
            except ValueError:
                pass
        return STRING_FORMAT

    def _apply_data_format(self, data: "any", fmt: str = None) -> "any":
        if fmt == ASDOT_FORMAT:
            part1 = int(data / 2**16)
            if part1 == 0:
                return str(data)
            part2 = data - (part1 * 2**16)
            return f"{part1}.{part2}"
        if fmt == STRING_FORMAT:
            return str(data)
        return data

    def _validate_format(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> str:
        # pylint: disable=too-many-branches
        if isinstance(data, int):
            # BGP AS number is in asplain format.
            int_data = data
            if self._format and self._format != ASPLAIN_FORMAT:
                self._raise_exceptions(
                    f"BGP AS number is not in {self._format} format", path, env
                )
            self._detected_format = ASPLAIN_FORMAT
        else:
            if self._format == ASPLAIN_FORMAT:
                self._raise_exceptions(
                    "BGP AS number is not in asplain format", path, env
                )

            if data.count(".") == 1:
                # BGP AS number is in asdot format.
                part1, part2 = data.split(".")
                try:
                    part1 = int(part1)
                    part2 = int(part2)
                except ValueError:
                    self._raise_exceptions("Invalid BGP AS format", path, env)
                int_data = part1 * 2**16 + part2
                if self._format and self._format != ASDOT_FORMAT:
                    self._raise_exceptions(
                        f"BGP AS number is not in {self._format} format", path, env
                    )
                self._detected_format = ASDOT_FORMAT
            else:
                # BGP AS number is a string
                if len(data) == 0:
                    self._raise_exceptions(
                        "Empty string is not allowed as a BGP AS number", path, env
                    )
                elif len(data) > 1 and data[0] == "0":
                    self._raise_exceptions(
                        "No leading zeros allowed in a BGP AS number", path, env
                    )
                try:
                    int_data = int(data)
                except ValueError:
                    self._raise_exceptions("Invalid BGP AS format", path, env)
                self._detected_format = STRING_FORMAT
                if int_data > 65535 and self._format and self._format == ASDOT_FORMAT:
                    self._raise_exceptions(
                        "BGP AS number is not in asdot format", path, env
                    )
        if not 0 <= int_data <= 4294967295:
            self._raise_exceptions(
                "BGP AS number is out of defined range 0-4294967295", path, env
            )
        return int_data

    def _preprocess_data(self, data: any, env: "ValEnv"):
        if CONVERT_DATA in env.flags:
            if self._convert_to_format is not None:
                data = self._apply_data_format(data, self._convert_to_format)
            else:
                data = self._apply_data_format(data, self._detected_format)
        else:
            data = self._apply_data_format(data, self._detected_format)
        return data

    def _test_data_type(self, data: str, path: str, env: "ValEnv" = None) -> "str":
        data = super()._test_data_type(data=data, path=path, env=env)
        data = self._validate_format(data=data, path=path, env=env)
        self._validate_scope(data=data, path=path, env=env)
        return data

    @property
    def restrictions(self) -> "list[str]":
        """
        Get all restrictions for valid data
        """
        ret_list = super().restrictions
        _restrictions = []
        if self._format:
            _restrictions.append((f"must be in format '{self._format}'"))
        if self._public is not None:
            _restrictions.append(f"must be {'public' if self._public else 'private'}")

        if _restrictions:
            ret_list.extend(
                ["BGP AS number:"]
                + [f"{self._markdown_indent}- {r}" for r in _restrictions]
            )
        return ret_list

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        return self._doc_data_type

    @property
    def markdown_data_type(self) -> str:
        """Get markdown data type string"""
        return f"<code>{self._doc_data_type}</code>"

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        allowed_formats = (
            "allowed_values:<br>"
            + f"{cls._markdown_indent}- {ASPLAIN_FORMAT}<br>"
            + f"{cls._markdown_indent}- {ASDOT_FORMAT}<br>"
            + f"{cls._markdown_indent}- {STRING_FORMAT}"
        )
        f_desc = (
            f"- {ASPLAIN_FORMAT} example: 100100<br>"
            + f"- {ASDOT_FORMAT} example: 100.100<br>"
            + f"- {STRING_FORMAT} example: '100100'"
        )
        return super()._doc_options_md_upper_part() + [
            f"| **format** | <code>enum</code> | | {allowed_formats} | {f_desc} |",
            "| **public** | <code>bool</code> | | | "
            + "BGP AS must be a public (true) or private (false) number |",
            f"| **convert_to_format** | <code>enum</code> | | {allowed_formats} | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            f"format: <{ASPLAIN_FORMAT}|{ASDOT_FORMAT}|{STRING_FORMAT}>",
            "public: <true|false>",
            f"convert_to_format: <{ASPLAIN_FORMAT}|{ASDOT_FORMAT}|{STRING_FORMAT}>",
        ]


DOC_DT_CLASS = BgpAs  # pylint: disable=invalid-name



"""
Mixins for minimal and maximal value/length
"""

# pylint: disable=too-few-public-methods

from nadap.errors import SchemaDefinitionError
from nadap.base import ValEnv


class _PopOptionsMixin:
    """
    Mixin for popping the options
    """

    def __init__(self, **kwargs):
        if not hasattr(self, "_minimum_default"):
            self._minimum_default = None
        self.minimum = None
        self._min_fmt = None
        if not hasattr(self, "_maximum_default"):
            self._maximum_default = None
        self.maximum = None
        self._max_fmt = None
        super().__init__(**kwargs)

    def _pop_options(self, definition: dict, schema_path: str):
        self.maximum = definition.pop("maximum", self._maximum_default)
        self.minimum = definition.pop("minimum", self._minimum_default)
        super()._pop_options(definition=definition, schema_path=schema_path)


class MinMaxValueMixin(_PopOptionsMixin):
    """
    Add min/max tests for value data
    """

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        """
        Check if data matches defined data type and apply preprocessing
        """
        data = super()._validate_data(data=data, path=path, env=env)
        if self.minimum is not None and data < self.minimum:
            self._create_finding_with_error(
                msg=f"Value is lower than defined minimum of {self.minimum}",
                path=path,
                env=env,
            )
        if self.maximum is not None and data > self.maximum:
            self._create_finding_with_error(
                msg=f"Value is greater than defined maximum of {self.maximum}",
                path=path,
                env=env,
            )
        return data

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        if self.maximum is not None:
            self._max_fmt = self._detect_data_format(self.maximum)
            self.maximum = self._test_data_type(
                data=self.maximum, path=f"{schema_path}.maximum"
            )
        if self.minimum is not None:
            self._min_fmt = self._detect_data_format(self.minimum)
            self.minimum = self._test_data_type(
                data=self.minimum, path=f"{schema_path}.minimum"
            )
        if (
            self.minimum is not None
            and self.maximum is not None
            and self.minimum > self.maximum
        ):
            raise SchemaDefinitionError(
                "maximum must be greater than or equal to minimum",
                schema_path,
            )

    @property
    def restrictions(self) -> "list[str]":
        """
        Get all restrictions for valid data
        """
        ret_list = super().restrictions
        if self.minimum is not None:
            ret_list.append(
                f"min: {self._apply_data_format(self.minimum, self._min_fmt)}"
            )
        if self.maximum is not None:
            ret_list.append(
                f"max: {self._apply_data_format(self.maximum, self._max_fmt)}"
            )
        return ret_list

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **minimum** | <code>"
            + "&#124;".join(sorted([x.__name__ for x in cls._cls_python_classes]))
            + "</code> | | |  Value must be greater or equal |",
            "| **maximum** | <code>"
            + "&#124;".join(sorted([x.__name__ for x in cls._cls_python_classes]))
            + "</code> | | >= 'minimum' |  Value must be lower or equal |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            f"maximum: <{'|'.join(sorted([x.__name__ for x in cls._cls_python_classes]))}>",
            f"minimum: <{'|'.join(sorted([x.__name__ for x in cls._cls_python_classes]))}>",
        ]


class MinMaxLengthMixin(_PopOptionsMixin):
    """
    Add min/max tests for value data length
    """

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        """
        Check if data matches defined data type and apply preprocessing
        """
        data = super()._validate_data(data=data, path=path, env=env)
        if self.minimum is not None and len(data) < self.minimum:
            self._create_finding_with_error(
                msg=f"Length is lower than defined minimum of {self.minimum}",
                path=path,
                env=env,
            )
        if self.maximum is not None and len(data) > self.maximum:
            self._create_finding_with_error(
                msg=f"Length is greater than defined maximum of {self.maximum}",
                path=path,
                env=env,
            )
        return data

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        if self.maximum is not None:
            if not isinstance(self.maximum, int):
                raise SchemaDefinitionError(
                    "Data is not an instance of 'int'",
                    f"{schema_path}.maximum",
                )
            if self.maximum < 0:
                raise SchemaDefinitionError(
                    "Must be >= 0",
                    f"{schema_path}.maximum",
                )
        if self.minimum is not None:
            if not isinstance(self.minimum, int):
                raise SchemaDefinitionError(
                    "Data is not an instance of 'int'",
                    f"{schema_path}.minimum",
                )
            if self.minimum < 0:
                raise SchemaDefinitionError(
                    "Must be >= 0",
                    f"{schema_path}.minimum",
                )
        if (
            self.minimum is not None
            and self.maximum is not None
            and self.minimum > self.maximum
        ):
            raise SchemaDefinitionError(
                "maximum must be greater than or equal to minimum",
                schema_path,
            )

    @property
    def restrictions(self) -> "list[str]":
        """
        Get all restrictions for valid data
        """
        ret_list = super().restrictions
        if self.minimum is not None:
            ret_list.append(f"min length: {self.minimum}")
        if self.maximum is not None:
            ret_list.append(f"max length: {self.maximum}")
        return ret_list

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **minimum** | <code>int</code> | | min: 1 "
            + "| Data's length must be greater or equal |",
            "| **maximum** | <code>int</code> | | min: 1<br>>= 'minimum' | "
            + "Data's length must be lower or equal |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "maximum: <int>",
            "minimum: <int>",
        ]



"""
Hostname data type class
"""

# pylint: disable=too-few-public-methods

import nadap.results
import nadap.types.base
import nadap.mixin.regex_mode
import nadap.mixin.min_max
import nadap.mixin.replace_empty_to
import nadap.references
import nadap.schema
from nadap.base import ValEnv, UNDEFINED
from nadap.errors import DataValidationError, SchemaDefinitionError

DOC_DT_NAME = "Hostname"
DOC_DT_DESCRIPTION = """
A **hostname** data type tests data for being an instance of
python's built-in class `str` and if it represents a hostname/FQDN.
"""
DOC_DT_FEATURES = """
- Validate conformity to IETF RFC 1123
- Validate string length
- Validate label count
- Validate against allowed and not allowed regex patterns
  (always full matching the string)
- Supports **Referencing Feature**
- Data type conversion is not supported
"""
DOC_DT_YAML_EXAMPLE = r"""
type: hostname
description: "Example hostname definition"
default_value: "my-host"

minimum: 3
maximum: 15
minimum_labels: 3
maximum_labels: 8

not_allowed_values:
  - nadap-.*
relax_length: false

reference: ref_key
"""


class Hostname(
    nadap.mixin.replace_empty_to.ReplaceEmptyToMixin,
    nadap.mixin.min_max.MinMaxLengthMixin,
    nadap.types.base.BaseType,
):
    """
    Hostname datatype class
    """

    data_type_name = "hostname"
    _cls_python_classes = [str]
    _convert_to_classes = {}
    _support_replace_empty_to = True
    _doc_data_type = "str"

    def __init__(self, **kwargs):
        self.allowed_values = None
        self.not_allowed_values = []
        self._relax_length = True
        self._min_labels = None
        self._max_labels = None
        super().__init__(**kwargs)

    def _validate_allowed_values_option(self, schema_path: str):
        if self.allowed_values is None:
            return
        regex_objects = []
        av_path = f"{schema_path}.allowed_values"
        nadap.schema.is_non_empty_list(self.allowed_values, av_path)
        for index, value in enumerate(self.allowed_values):
            av_i_path = f"{av_path}[{index}]"
            self._test_data_type(value, av_i_path)
            regex_objects.append(
                nadap.schema.compile_regex_string(
                    pattern=value,
                    multiline=False,
                    fullmatch=True,
                    schema_path=av_i_path,
                )
            )
        self.allowed_values = regex_objects

    def _validate_not_allowed_values_option(self, schema_path: str):
        regex_objects = []
        av_path = f"{schema_path}.not_allowed_values"
        nadap.schema.is_list(self.not_allowed_values, av_path)
        for index, value in enumerate(self.not_allowed_values):
            av_i_path = f"{av_path}[{index}]"
            self._test_data_type(value, av_i_path)
            regex_objects.append(
                nadap.schema.compile_regex_string(
                    pattern=value,
                    multiline=False,
                    fullmatch=True,
                    schema_path=av_i_path,
                )
            )
        self.not_allowed_values = regex_objects

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        self._validate_allowed_values_option(schema_path)
        self._validate_not_allowed_values_option(schema_path)
        if self.maximum is not None:
            if self._relax_length:
                if self.maximum > 255:
                    raise SchemaDefinitionError(
                        msg="Exceeds allowed hostname length of 255 characters",
                        path=f"{schema_path}.maximum",
                    )
            elif self.maximum > 63:
                raise SchemaDefinitionError(
                    msg="Exceeds allowed hostname length of 63 characters",
                    path=f"{schema_path}.maximum",
                )
        if self._min_labels is not None:
            min_path = f"{schema_path}.minimum_labels"
            nadap.schema.is_int(self._min_labels, min_path)
            if self._min_labels < 1:
                raise SchemaDefinitionError(msg="Must be > 0", path=min_path)
        if self._max_labels is not None:
            max_path = f"{schema_path}.maximum_labels"
            nadap.schema.is_int(self._max_labels, max_path)
            if self._max_labels < 1:
                raise SchemaDefinitionError(msg="Must be > 0", path=max_path)
            if self._min_labels is not None and self._min_labels > self._max_labels:
                raise SchemaDefinitionError(
                    msg="Must be >= 'minimum_labels'", path=max_path
                )

    def _pop_options(self, definition: dict, schema_path: str):
        self.allowed_values = definition.pop("allowed_values", None)
        self.not_allowed_values = definition.pop("not_allowed_values", [])
        self._relax_length = definition.pop("relax_length", True)
        self._min_labels = definition.pop("minimum_labels", None)
        self._max_labels = definition.pop("maximum_labels", None)
        super()._pop_options(definition, schema_path)

    def _convert_reference_data(self, data: any) -> any:
        """
        Convert data to reference data
        """
        return data.lower()

    def _validate_allowed_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        if not self.allowed_values:
            return
        if not nadap.schema.match_regex_objects(self.allowed_values, data)[0]:
            self._create_finding_with_error(
                msg="Data does not match any allowed regex patterns",
                path=path,
                env=env,
            )

    def _validate_not_allowed_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        if not self.not_allowed_values:
            return
        matches, pattern = nadap.schema.match_regex_objects(
            self.not_allowed_values, data
        )
        if matches:
            self._create_finding_with_error(
                msg=f"Data does match not allowed regex pattern '{pattern.pattern}'",
                path=path,
                env=env,
            )

    def _validate_rfc_length(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        invalid = False
        if self.maximum is None:
            if self._relax_length:
                if len(data) > 255:
                    env.findings.append(
                        nadap.results.ValidationFinding(
                            message="String length exceeds allowed 255 characters",
                            path=path,
                        )
                    )
                    invalid = True
            elif len(data) > 63:
                env.findings.append(
                    nadap.results.ValidationFinding(
                        message="String length exceeds allowed 63 characters",
                        path=path,
                    )
                )
                invalid = True
        if invalid:
            raise DataValidationError()
        if self._replace_empty_to is UNDEFINED and not data:
            self._create_finding_with_error(
                msg="Hostname must have at least one character",
                path=path,
                env=env,
            )

    def _validate_label(
        self,
        label: str,
        path: str,
        env: "ValEnv",
        label_index: int,
    ):
        messages = []
        for char_index, char_ in enumerate(label):
            if char_index == 0 and not char_.isalnum():
                if label_index > 0:
                    messages.append(
                        f"Hostname's label {label_index} does not start "
                        + "with an alphanumeric character"
                    )
                else:
                    messages.append(
                        "Hostname does not start with an alphanumeric character"
                    )
            elif char_index == len(label) - 1 and not char_.isalnum():
                if label_index > 0:
                    messages.append(
                        f"Hostname's label {label_index} does not end "
                        + "with an alphanumeric character"
                    )
                else:
                    messages.append(
                        "Hostname does not end with an alphanumeric character"
                    )
            elif not char_.isalnum() and not char_ == "-":
                if label_index > 0:
                    messages.append(
                        f"Hostname's label {label_index} contains illegal "
                        + f"character at position {label_index + 1}"
                    )
                else:
                    messages.append(
                        f"Hostname contains illegal character at position {char_index + 1}"
                    )
        if messages:
            for msg in messages:
                env.findings.append(
                    nadap.results.ValidationFinding(
                        message=msg,
                        path=path,
                    )
                )
            raise DataValidationError()

    def _validate_rfc_labels(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        message = None
        invalid = False
        labels = data.split(".")
        if self._min_labels is not None and len(labels) < self._min_labels:
            message = f"Label count is lower than defined minimum of {self._min_labels}"
        if self._max_labels is not None and len(labels) > self._max_labels:
            message = (
                f"Label count is greater than defined maximum of {self._max_labels}"
            )
        for label_index, label in enumerate(labels, start=1):
            try:
                self._validate_label(
                    label=label,
                    path=path,
                    env=env,
                    label_index=label_index if len(labels) > 1 else 0,
                )
            except DataValidationError:
                invalid = True
        if invalid:
            raise DataValidationError()
        if message:
            env.findings.append(
                nadap.results.ValidationFinding(
                    message=message,
                    path=path,
                )
            )
            raise DataValidationError()

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        data = super()._validate_data(data=data, path=path, env=env)
        invalid = False
        try:
            self._validate_rfc_length(data=data, path=path, env=env)
        except DataValidationError:
            invalid = True
        try:
            self._validate_rfc_labels(data=data, path=path, env=env)
        except DataValidationError:
            invalid = True
        if invalid:
            raise DataValidationError()
        self._validate_allowed_data(data=data, path=path, env=env)
        self._validate_not_allowed_data(data=data, path=path, env=env)
        return data

    @property
    def restrictions(self) -> "list[str]":
        """
        Get all restrictions for valid data
        """
        ret_list = super().restrictions
        if self.maximum is None and not self._relax_length:
            ret_list.append("max length: 63")
        if self._min_labels is not None:
            ret_list.append(f"min labels: {self._min_labels}")
        if self._max_labels is not None:
            ret_list.append(f"max labels: {self._max_labels}")
        if self.allowed_values:
            ret_list.append("allowed regex patterns:")
            ret_list.extend([rf" - `{x.pattern.pattern}`" for x in self.allowed_values])
        if self.not_allowed_values:
            ret_list.append("not allowed regex patterns:")
            ret_list.extend(
                [rf" - `{x.pattern.pattern}`" for x in self.not_allowed_values]
            )
        return ret_list

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        return self.data_type_name.title()

    @property
    def markdown_data_type(self) -> str:
        """Get markdown data type string"""
        return f"<code>{self.data_type_name.title()}</code>"

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **relax_length** | <code>bool</code> | <code>True</code> | | "
            + "Relax hostname length from 63 to 255 characters |",
            "| **minimum_labels** | <code>int</code> | | min: 1 | "
            + "Minimum amount of labels within in hostname |",
            "| **maximum_labels** | <code>int</code> | | min: 1<br>>= 'minimum_labels' | "
            + "Maximum amount of labels within in hostname |",
            "| **allowed_values** | <code>list[str]</code> | | | "
            + " Data must match one of these regex patterns |",
            "| &nbsp;&nbsp;- < str > | <code>str</code> | | | |",
            "| **not_allowed_values** | <code>list[str]</code> | | | "
            + " Data mustn't match any of these regex patterns |",
            "| &nbsp;&nbsp;- < str > | <code>str</code> | | | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "relax_length: <true|false>",
            "minimum_labels: <int>",
            "maximum_labels: <int>",
            "allowed_values:",
            "  - <str>",
            "not_allowed_values:",
            "  - <str>",
        ]


DOC_DT_CLASS = Hostname  # pylint: disable=invalid-name



"""
Mixin for IP data types
"""

# pylint: disable=too-few-public-methods

import ipaddress

from nadap.base import ValEnv
from nadap.errors import SchemaDefinitionError
import nadap.mixin.allowed_value
import nadap.mixin.not_allowed_value


class IpMixin(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
):
    """
    Add handling of strings representing IPv4 or IPv6 string
    """

    _cls_python_classes = [str]
    _doc_data_type = "str"
    _ip_data_type_name = "IP"

    def __init__(self, **kwargs):
        self._test_loopback = None
        self._test_link_local = None
        self._test_multicast = None
        self._test_private = None
        self._test_public = None
        super().__init__(**kwargs)

    def _get_ip_obj(self, data):
        return ipaddress.ip_address(data)

    def _test_data_type(self, data: any, path: str, env: "ValEnv" = None) -> "any":
        # pylint: disable=raise-missing-from
        data = super()._test_data_type(data=data, path=path, env=env)
        try:
            self._get_ip_obj(data)
        except ValueError:
            msg = f"String does not represent an {self._ip_data_type_name}"
            if env is None:
                raise SchemaDefinitionError(
                    msg=msg,
                    path=path,
                )
            self._create_finding_with_error(msg=msg, path=path, env=env)
        return data

    def _pop_options(self, definition: dict, schema_path: str):
        self._test_loopback = definition.pop("loopback", None)
        self._test_link_local = definition.pop("link_local", None)
        self._test_private = definition.pop("private", None)
        self._test_public = definition.pop("public", None)
        self._test_multicast = definition.pop("multicast", None)
        super()._pop_options(definition=definition, schema_path=schema_path)

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        if (
            self._test_link_local is not None
            and self._test_loopback is not None
            and self._test_link_local == self._test_loopback
        ):
            raise SchemaDefinitionError(
                "Mismatching test for link_local and loopback", schema_path
            )
        if (
            self._test_link_local is not None
            and self._test_multicast is not None
            and self._test_link_local == self._test_multicast
        ):
            raise SchemaDefinitionError(
                "Mismatching test for link_local and multicast", schema_path
            )
        if (
            self._test_loopback is not None
            and self._test_multicast is not None
            and self._test_loopback == self._test_multicast
        ):
            raise SchemaDefinitionError(
                "Mismatching test for loopback and multicast", schema_path
            )
        if (
            self._test_private is not None
            and self._test_public is not None
            and self._test_private == self._test_public
        ):
            raise SchemaDefinitionError(
                "Mismatching test for private and public", schema_path
            )

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        """
        Check if data matches IP options
        """
        data = super()._validate_data(data=data, path=path, env=env)
        ip_obj = self._get_ip_obj(data)
        msg = None
        if (
            self._test_link_local is not None
            and self._test_link_local != ip_obj.is_link_local
        ):
            msg = "Is link-local" if ip_obj.is_link_local else "Is not link-local"
        elif (
            self._test_loopback is not None
            and self._test_loopback != ip_obj.is_loopback
        ):
            msg = "Is loopback" if ip_obj.is_loopback else "Is not loopback"
        elif (
            self._test_multicast is not None
            and self._test_multicast != ip_obj.is_multicast
        ):
            msg = "Is multicast" if ip_obj.is_multicast else "Is not multicast"
        elif self._test_private is not None and self._test_private != ip_obj.is_private:
            msg = "Is private" if ip_obj.is_private else "Is not private"
        elif self._test_public is not None and self._test_public != ip_obj.is_global:
            msg = "Is public" if ip_obj.is_global else "Is not public"
        if msg:
            self._create_finding_with_error(msg=msg, path=path, env=env)

        return data

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        return self._ip_data_type_name

    @property
    def markdown_data_type(self) -> str:
        """Get markdown data type string"""
        return f"<code>{self._ip_data_type_name}</code>"

    @property
    def restrictions(self) -> "list[str]":
        """
        Get all restrictions for valid data
        """
        ret_list = super().restrictions
        opt_list = []
        if self._test_link_local is not None:
            opt_list.append(
                f"{self._markdown_indent}- link-local: {self._test_link_local}"
            )
        if self._test_loopback is not None:
            opt_list.append(f"{self._markdown_indent}- loopback: {self._test_loopback}")
        if self._test_multicast is not None:
            opt_list.append(
                f"{self._markdown_indent}- multicast: {self._test_multicast}"
            )
        if self._test_private is not None:
            opt_list.append(f"{self._markdown_indent}- private: {self._test_private}")
        if self._test_public is not None:
            opt_list.append(f"{self._markdown_indent}- public: {self._test_public}")
        if opt_list:
            ret_list.append("IP restrictions:")
            ret_list.extend(opt_list)
        return ret_list

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **link_local** | <code>bool</code> | | | IP must be link-local |",
            "| **loopback** | <code>bool</code> | | | IP must be loopback |",
            "| **multicast** | <code>bool</code> | | | IP must be multicast |",
            "| **private** | <code>bool</code> | | | IP must be private |",
            "| **public** | <code>bool</code> | | | IP must be public/global |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "",
            "link_local: <true|false",
            "loopback: <true|false",
            "multicast: <true|false",
            "private: <true|false",
            "public: <true|false",
        ]



"""
Enum type class
"""

# pylint: disable=too-few-public-methods

import nadap.results
import nadap.schema
import nadap.types.base
import nadap.mixin.allowed_value

DOC_DT_NAME = "Enum"
DOC_DT_DESCRIPTION = """
A **enum** data type represents a selection of allowed discrete data values.
"""
DOC_DT_FEATURES = """
- Supports **Referencing Feature**. Adds an option to lookup values in other namespaces.
- Data type conversion is not supported
"""
DOC_DT_YAML_EXAMPLE = """
type: enum
description: "Example enum definition"
default_value: 1

allowed_values:
  - key1: 1
    key2: string2
  - 1
  - 2
  -
    - nadap
    - rulez!

reference:
  key: ref_key
  mode: producer
"""


class Enum(nadap.mixin.allowed_value.AllowedValueMixin, nadap.types.base.BaseType):
    """
    Enum datatype class
    Only discrete values are allowed.
    """

    data_type_name = "enum"
    _convert_to_classes = {}


DOC_DT_CLASS = Enum  # pylint: disable=invalid-name



"""
Central import for types
"""

from nadap.nadap import Nadap
from nadap.errors import SchemaDefinitionError, DataValidationError
from nadap.schema import Schema
from nadap.references import ReferenceFinding
from nadap.results import ValidationFinding
from nadap.base import SET_DEFAULTS, CONVERT_DATA, NOFLAG


__version__ = "1.5.2"
version_info = (1, 5, 2)



"""
Integer data type classes
"""

# pylint: disable=too-few-public-methods

import nadap.schema
import nadap.types.base
import nadap.mixin.accuracy
import nadap.mixin.not_allowed_value
import nadap.mixin.min_max
import nadap.mixin.ranges
import nadap.mixin.allowed_value

DOC_DT_NAME = "Integer"
DOC_DT_DESCRIPTION = """
An **int** data type tests data for being an instance of
python's built-in class `int`.

> If using *int* data type with allowed values and no conversion defined,
> check *enum* data type as a better option.
"""
DOC_DT_FEATURES = f"""
- Validate minimal and maximal integer value{nadap.mixin.ranges.DOC_FEATURES}
- Validate against allowed and not allowed integer values
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: int
description: "Example int definition"

minimum: 1
maximum: 4096

not_allowed_values:
  - 1001
  - 1002

reference:
  key: ref_key
  mode: consumer
"""


class Int(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.mixin.min_max.MinMaxValueMixin,
    nadap.types.base.BaseType,
):
    """
    Integer datatype class
    """

    data_type_name = "int"
    _cls_python_classes = [int]
    _convert_to_classes = {"str": str, "float": float}
    _doc_data_type = "int"


DOC_DT_CLASS = Int  # pylint: disable=invalid-name
